import json  # ujson causes issues
import logging

from .stream import (
    MarketStream,
    OrderStream,
)

logger = logging.getLogger(__name__)


class BaseListener(object):

    def __init__(self, max_latency=0.5):
        self.max_latency = max_latency

        self.connection_id = None
        self.stream = None
        self.stream_type = None  # marketSubscription/orderSubscription
        self.stream_unique_id = None

    def register_stream(self, unique_id, operation):
        if self.stream is not None:
            logger.warning('[Listener: %s]: stream already registered, replacing data' % unique_id)
        self.stream_unique_id = unique_id
        self.stream_type = operation
        self.stream = self._add_stream(unique_id, operation)

    def on_data(self, raw_data):
        logger.info(raw_data)

    def snap(self, market_ids=None):
        """Returns a 'snap' of the current cache
        data.

        :param list market_ids: Market ids to return
        :return: Return List of resources
        """
        if self.stream:
            return self.stream.snap(market_ids)
        else:
            return []

    @property
    def updates_processed(self):
        if self.stream:
            return self.stream._updates_processed

    @property
    def initial_clk(self):
        if self.stream is not None:
            return self.stream._initial_clk

    @property
    def clk(self):
        if self.stream is not None:
            return self.stream._clk

    def _add_stream(self, unique_id, operation):
        logger.info('Register: %s %s' % (operation, unique_id))

    def __str__(self):
        return 'BaseListener'

    def __repr__(self):
        return '<BaseListener>'


class StreamListener(BaseListener):
    """Stream listener, processes results from socket,
    holds a stream which can hold order or market book
    caches
    """

    def __init__(self, output_queue=None, max_latency=0.5, lightweight=False):
        """
        :param Queue output_queue: Queue used to return data
        :param float max_latency: Logs warning if latency above value
        :param bool lightweight: Returns dict instead of resource
        """
        super(StreamListener, self).__init__(max_latency)
        self.output_queue = output_queue
        self.lightweight = lightweight

    def on_data(self, raw_data):
        """Called when raw data is received from connection.
        Override this method if you wish to manually handle
        the stream data

        :param raw_data: Received raw data
        :return: Return False to stop stream and close connection
        """
        try:
            data = json.loads(raw_data)
        except ValueError:
            logger.error('value error: %s' % raw_data)
            return

        unique_id = data.get('id')

        if self._error_handler(data, unique_id):
            return False

        operation = data['op']
        if operation == 'connection':
            self._on_connection(data, unique_id)
        elif operation == 'status':
            self._on_status(data, unique_id)
        elif operation in ['mcm', 'ocm']:
            # historic data does not contain unique_id
            if self.stream_unique_id not in [unique_id, 'HISTORICAL']:
                logger.warning('Unwanted data received from uniqueId: %s, expecting: %s' %
                               (unique_id, self.stream_unique_id))
                return
            self._on_change_message(data, unique_id)

    def _on_connection(self, data, unique_id):
        """Called on collection operation

        :param data: Received data
        """
        if unique_id is None:
            unique_id = self.stream_unique_id
        self.connection_id = data.get('connectionId')
        logger.info('[Connect: %s]: connection_id: %s' % (unique_id, self.connection_id))

    @staticmethod
    def _on_status(data, unique_id):
        """Called on status operation

        :param data: Received data
        """
        status_code = data.get('statusCode')
        logger.info('[Subscription: %s]: %s' % (unique_id, status_code))

    def _on_change_message(self, data, unique_id):
        change_type = data.get('ct', 'UPDATE')

        logger.debug('[Subscription: %s]: %s: %s' % (unique_id, change_type, data))

        if change_type == 'SUB_IMAGE':
            self.stream.on_subscribe(data)
        elif change_type == 'RESUB_DELTA':
            self.stream.on_resubscribe(data)
        elif change_type == 'HEARTBEAT':
            self.stream.on_heartbeat(data)
        elif change_type == 'UPDATE':
            self.stream.on_update(data)

    def _add_stream(self, unique_id, stream_type):
        if stream_type == 'marketSubscription':
            return MarketStream(self)
        elif stream_type == 'orderSubscription':
            return OrderStream(self)

    @staticmethod
    def _error_handler(data, unique_id):
        """Called when data first received

        :param data: Received data
        :param unique_id: Unique id
        :return: True if error present
        """
        if data.get('statusCode') == 'FAILURE':
            logger.error('[Subscription: %s] %s: %s' % (unique_id, data.get('errorCode'), data.get('errorMessage')))
            if data.get('connectionClosed'):
                return True
        if data.get('status'):
            # Clients shouldn't disconnect if status 503 is returned; when the stream
            # recovers updates will be sent containing the latest data
            logger.warning('[Subscription: %s] status: %s' % (unique_id, data['status']))

    def __str__(self):
        return 'StreamListener'

    def __repr__(self):
        return '<StreamListener>'



from .baseendpoint import BaseEndpoint
from .. import resources
from ..utils import clean_locals
from ..filters import time_range


class Account(BaseEndpoint):
    """
    Account operations.
    """

    URI = 'AccountAPING/v1.0/'
    connect_timeout = 6.05

    def get_account_funds(self, wallet=None, session=None, lightweight=None):
        """
        Get available to bet amount.

        :param str wallet: Name of the wallet in question
        :param requests.session session: Requests session object
        :param bool lightweight: If True will return dict not a resource

        :rtype: resources.AccountFunds
        """
        params = clean_locals(locals())
        method = '%s%s' % (self.URI, 'getAccountFunds')
        (response, elapsed_time) = self.request(method, params, session)
        return self.process_response(response, resources.AccountFunds, elapsed_time, lightweight)

    def get_account_details(self, session=None, lightweight=None):
        """
        Returns the details relating your account, including your discount
        rate and Betfair point balance.

        :param requests.session session: Requests session object
        :param bool lightweight: If True will return dict not a resource

        :rtype: resources.AccountDetails
        """
        params = clean_locals(locals())
        method = '%s%s' % (self.URI, 'getAccountDetails')
        (response, elapsed_time) = self.request(method, params, session)
        return self.process_response(response, resources.AccountDetails, elapsed_time, lightweight)

    def get_account_statement(self, locale=None, from_record=None, record_count=None, item_date_range=time_range(),
                              include_item=None, wallet=None, session=None, lightweight=None):
        """
        Get account statement.

        :param str locale: The language to be used where applicable.
        :param int from_record: Specifies the first record that will be returned
        :param int record_count: Specifies the maximum number of records to be returned.
        :param dict item_date_range: Return items with an itemDate within this date range.
        :param str include_item: Which items to include, if not specified then defaults to ALL.
        :param str wallet: Which wallet to return statementItems for.
        :param requests.session session: Requests session object
        :param bool lightweight: If True will return dict not a resource

        :rtype: resources.AccountStatementResult
        """
        params = clean_locals(locals())
        method = '%s%s' % (self.URI, 'getAccountStatement')
        (response, elapsed_time) = self.request(method, params, session)
        return self.process_response(response, resources.AccountStatementResult, elapsed_time, lightweight)

    def list_currency_rates(self, from_currency=None, session=None, lightweight=None):
        """
        Returns a list of currency rates based on given currency

        :param str from_currency: The currency from which the rates are computed
        :param requests.session session: Requests session object
        :param bool lightweight: If True will return dict not a resource

        :rtype: list[resources.CurrencyRate]
        """
        params = clean_locals(locals())
        method = '%s%s' % (self.URI, 'listCurrencyRates')
        (response, elapsed_time) = self.request(method, params, session)
        return self.process_response(response, resources.CurrencyRate, elapsed_time, lightweight)

    def transfer_funds(self, session=None):
        """
        Transfer funds between the UK Exchange and other wallets

        :param requests.session session: Requests session object

        :rtype: resources.TransferFunds
        """
        raise DeprecationWarning(
            'As of 20/09/2016 AUS wallet has been removed, function still available for when '
            'accounts are added in 2017.')

    @property
    def url(self):
        return '%s%s' % (self.client.api_uri, 'account/json-rpc/v1')



import unittest

from betfairlightweight.filters import (
    market_filter,
    time_range,
    price_data,
    ex_best_offers_overrides,
    price_projection,
    place_instruction,
    limit_on_close_order,
    limit_order,
    cancel_instruction,
    market_on_close_order,
    replace_instruction,
    update_instruction,
    streaming_market_data_filter,
    streaming_market_filter,
    streaming_order_filter,
)


class FilterTest(unittest.TestCase):

    def test_streaming_market_filter(self):
        response = streaming_market_filter()
        assert response == {}

        response = streaming_market_filter(market_ids=[1, 2])
        assert response == {'marketIds': [1, 2]}

    def test_streaming_market_data_filter(self):
        response = streaming_market_data_filter()
        assert response == {}

        response = streaming_market_data_filter(ladder_levels=3)
        assert response == {'ladderLevels': 3}

    def test_streaming_order_filter(self):
        response = streaming_order_filter()
        assert response == {}

        response = streaming_order_filter(include_overall_position=True)
        assert response == {'includeOverallPosition': True}

    def test_time_range(self):
        response = time_range()
        assert response == {'from': None, 'to': None}

        response = time_range(from_='123', to='456')
        assert response == {'from': '123', 'to': '456'}

    def test_market_filter(self):
        response = market_filter()
        assert response == {}

        response = market_filter(market_ids=['1.123'])
        assert response == {'marketIds': ['1.123']}

    def test_price_data(self):
        response = price_data()
        assert response == []

        response = price_data(ex_best_offers=True)
        assert response == ['EX_BEST_OFFERS']

    def test_ex_best_offers_overrides(self):
        response = ex_best_offers_overrides()
        assert response == {}

    def test_price_projection(self):
        response = price_projection()
        assert response == {
            'rolloverStakes': False, 'priceData': [], 'exBestOffersOverrides': {}, 'virtualise': True
        }

    def test_place_instruction(self):
        response = place_instruction('LIMIT', 123, 'LAY')
        assert response == {'orderType': 'LIMIT', 'selectionId': 123, 'side': 'LAY'}

    def test_limit_order(self):
        response = limit_order(size=1.1, price=123, persistence_type='LAPSE')
        assert response == {'size': 1.1, 'price': 123, 'persistenceType': 'LAPSE'}

    def test_limit_on_close_order(self):
        response = limit_on_close_order(1.1, 2.2)
        assert response == {'liability': 1.1, 'price': 2.2}

    def test_market_on_close_order(self):
        response = market_on_close_order(1.1)
        assert response == {'liability': 1.1}

    def test_cancel_instruction(self):
        response = cancel_instruction('1.123')
        assert response == {'betId': '1.123'}

    def test_replace_instruction(self):
        response = replace_instruction('1.123', 12)
        assert response == {'betId': '1.123', 'newPrice': 12}

    def test_update_instruction(self):
        response = update_instruction('1.123', 'LAPSE')
        assert response == {'betId': '1.123', 'newPersistenceType': 'LAPSE'}



from enum import Enum


GENERIC_JSON_RPC_EXCEPTIONS = {
    -32700: 'Invalid JSON was received by the server. An error occurred on the server while parsing the JSON text.',
    -32601: 'Method not found',
    -32602: 'Problem parsing the parameters, or a mandatory parameter was not found',
    -32603: 'Internal JSON-RPC error'
}


class RaceStatusEnum(Enum):
    DORMANT = 'There is no data available for this race.'
    DELAYED = 'The start of the race has been delayed'
    PARADING = 'The horses are in the parade ring'
    GOINGDOWN = 'The horses are going down to the starting post'
    GOINGBEHIND = 'The horses are going behind the stalls'
    ATTHEPOST = 'The horses are at the post'
    UNDERORDERS = 'The horses are loaded into the stalls/race is about to start'
    OFF = 'The race has started'
    FINISHED = 'The race has finished'
    FALSESTART = 'There has been a false start'
    PHOTOGRAPH = 'The result of the race is subject to a photo finish'
    RESULT = 'The result of the race has been announced'
    WEIGHEDIN = 'The jockeys have weighed in'
    RACEVOID = 'The race has been declared void'
    ABANDONED = 'The meeting has been cancelled'


class LoginExceptions(Enum):
    INVALID_USERNAME_OR_PASSWORD = "The username or password are invalid"
    ACCOUNT_NOW_LOCKED = "The account was just locked"
    ACCOUNT_ALREADY_LOCKED = "The account is already locked"
    PENDING_AUTH = "Pending authentication"
    TELBET_TERMS_CONDITIONS_NA = "Telbet terms and conditions rejected"
    DUPLICATE_CARDS = "Duplicate cards"
    SECURITY_QUESTION_WRONG_3X = "The user has entered wrong the security answer 3 times"
    KYC_SUSPEND = "KYC suspended"
    SUSPENDED = "The account is suspended"
    CLOSED = "The account is closed"
    SELF_EXCLUDED = "The account has been self-excluded"
    INVALID_CONNECTIVITY_TO_REGULATOR_DK = "The DK regulator cannot be accessed due to some internal problems in the " \
                                           "system behind or in at regulator; timeout cases included."
    NOT_AUTHORIZED_BY_REGULATOR_DK = "The user identified by the given credentials is not authorized in the DKs " \
                                     "jurisdictions due to the regulators policies. Ex = the user for which " \
                                     "this session should be created is not allowed to act(play bet) in the DKs " \
                                     "jurisdiction."
    INVALID_CONNECTIVITY_TO_REGULATOR_IT = "The IT regulator cannot be accessed due to some internal problems in the " \
                                           "system behind or in at regulator; timeout cases included."
    NOT_AUTHORIZED_BY_REGULATOR_IT = "The user identified by the given credentials is not authorized in the ITs " \
                                     "jurisdictions due to the regulators policies. Ex = the user for which this " \
                                     "session should be created is not allowed to act(play bet) in the ITs " \
                                     "jurisdiction."
    SECURITY_RESTRICTED_LOCATION = "The account is restricted due to security concerns"
    BETTING_RESTRICTED_LOCATION = "The account is accessed from a location where betting is restricted"
    TRADING_MASTER = "Trading Master Account"
    TRADING_MASTER_SUSPENDED = "Suspended Trading Master Account"
    AGENT_CLIENT_MASTER = "Agent Client Master"
    AGENT_CLIENT_MASTER_SUSPENDED = "Suspended Agent Client Master"
    DANISH_AUTHORIZATION_REQUIRED = "Danish authorization required"
    SPAIN_MIGRATION_REQUIRED = "Spain migration required"
    DENMARK_MIGRATION_REQUIRED = "Denmark migration required"
    SPANISH_TERMS_ACCEPTANCE_REQUIRED = "The latest Spanish terms and conditions version must be accepted"
    ITALIAN_CONTRACT_ACCEPTANCE_REQUIRED = "The latest Italian contract version must be accepted"
    CERT_AUTH_REQUIRED = "Certificate required or certificate present but could not authenticate with it"
    CHANGE_PASSWORD_REQUIRED = "Change password required"
    PERSONAL_MESSAGE_REQUIRED = "Personal message required for the user"
    INTERNATIONAL_TERMS_ACCEPTANCE_REQUIRE = "The latest international terms and conditions must be accepted prior " \
                                             "to logging in."
    EMAIL_LOGIN_NOT_ALLOWED = "This account has not opted in to log in with the email"
    MULTIPLE_USERS_WITH_SAME_CREDENTIAL = "There is more than one account with the same credential"
    ACCOUNT_PENDING_PASSWORD_CHANGE = "The account must undergo password recovery to reactivate"
    TEMPORARY_BAN_TOO_MANY_REQUEST = "The limit for successful login requests per minute has been exceeded. New " \
                                     "login attempts will be banned for 20 minutes"


class ApingException(Enum):
    TOO_MUCH_DATA = "The operation requested too much data exceeding the Market Data Request Limits."
    INVALID_INPUT_DATA = "The data input is invalid. A specific description is returned via errorDetails as shown " \
                         "below."
    INVALID_SESSION_INFORMATION = "The session token hasnt been provided is invalid or has expired."
    NO_APP_KEY = "An application key header (X-Application) has not been provided in the request"
    NO_SESSION = "A session token header (X-Authentication) has not been provided in the request"
    UNEXPECTED_ERROR = "An unexpected internal error occurred that prevented successful request processing."
    INVALID_APP_KEY = "The application key passed is invalid or is not present"
    TOO_MANY_REQUESTS = "There are too many pending requests e.g. a listMarketBook with Order/Match projections is " \
                        "limited to 3 concurrent requests. The error also applies to listCurrentOrders " \
                        "listMarketProfitAndLoss and listClearedOrders if you have 3 or more requests currently " \
                        "in execution"
    SERVICE_BUSY = "The service is currently too busy to service this request."
    TIMEOUT_ERROR = "The Internal call to downstream service timed out. Please note = If a TIMEOUT_ERROR error " \
                    "occurs on a placeOrders/replaceOrders request you should check listCurrentOrders to verify the " \
                    "status of your bets before placing further orders. Please allow up to 2 minutes for timed out " \
                    "order to appear."
    REQUEST_SIZE_EXCEEDS_LIMIT = "The request exceeds the request size limit. Requests are limited to a total of 250 " \
                                 "betId's/marketId's (or a combination of both)."
    ACCESS_DENIED = "The calling client is not permitted to perform the specific action e.g. the using a Delayed " \
                    "App Key when placing bets or attempting to place a bet from a restricted jurisdiction."


class MarketStatus(Enum):
    INACTIVE = "The market has been created but isn't yet available."
    OPEN = "The market is open for betting."
    SUSPENDED = "The market is suspended and not available for betting."
    CLOSED = "The market has been settled and is no longer available for betting."


class InstructionReportStatus(Enum):
    SUCCESS = ''
    FAILURE = ''
    TIMEOUT = ''


class InstructionReportErrorCode(Enum):
    INVALID_BET_SIZE = "bet size is invalid for your currency or your regulator"
    INVALID_RUNNER = "Runner does not exist includes vacant traps in greyhound racing"
    BET_TAKEN_OR_LAPSED = "Bet cannot be cancelled or modified as it has already been taken or has lapsed Includes " \
                          "attempts to cancel/modify market on close BSP bets and cancelling limit on close BSP bets"
    BET_IN_PROGRESS = "No result was received from the matcher in a timeout configured for the system"
    RUNNER_REMOVED = "Runner has been removed from the event"
    MARKET_NOT_OPEN_FOR_BETTING = "Attempt to edit a bet on a market that has closed."
    LOSS_LIMIT_EXCEEDED = "The action has caused the account to exceed the self imposed loss limit"
    MARKET_NOT_OPEN_FOR_BSP_BETTING = "Market now closed to bsp betting. Turned in-play or has been reconciled"
    INVALID_PRICE_EDIT = "Attempt to edit down the price of a bsp limit on close lay bet or edit up the price of a " \
                         "limit on close back bet"
    INVALID_ODDS = "Odds not on price ladder - either edit or placement"
    INSUFFICIENT_FUNDS = "Insufficient funds available to cover the bet action. Either the exposure limit or " \
                         "available to bet limit would be exceeded"
    INVALID_PERSISTENCE_TYPE = "Invalid persistence type for this market e.g. KEEP for a non bsp market"
    ERROR_IN_MATCHER = "A problem with the matcher prevented this action completing successfully"
    INVALID_BACK_LAY_COMBINATION = "The order contains a back and a lay for the same runner at overlapping prices. " \
                                   "This would guarantee a self match. This also applies to BSP limit on close bets"
    ERROR_IN_ORDER = "The action failed because the parent order failed"
    INVALID_BID_TYPE = "Bid type is mandatory"
    INVALID_BET_ID = "Bet for id supplied has not been found"
    CANCELLED_NOT_PLACED = "Bet cancelled but replacement bet was not placed"
    RELATED_ACTION_FAILED = "Action failed due to the failure of a action on which this action is dependent"
    NO_ACTION_REQUIRED = "the action does not result in any state change. eg changing a persistence to it's " \
                         "current value"


class ExecutionReportStatus(Enum):
    SUCCESS = "Order processed successfully"
    FAILURE = "Order failed."
    PROCESSED_WITH_ERRORS = "The order itself has been accepted but at least one (possibly all) actions have " \
                            "generated errors. This error only occurs for replaceOrders cancelOrders and " \
                            "updateOrders operations. The placeOrders operation will not return " \
                            "PROCESSED_WITH_ERRORS status as it is an atomic operation."
    TIMEOUT = "Order timed out."


class ExecutionReportErrorCode(Enum):
    ERROR_IN_MATCHER = "The matcher is not healthy"
    PROCESSED_WITH_ERRORS = "The order itself has been accepted but at least one (possibly all) actions have " \
                            "generated errors"
    BET_ACTION_ERROR = "There is an error with an action that has caused the entire order to be rejected. Check " \
                       "the instructionReports errorCode for the reason for the rejection of the order."
    INVALID_ACCOUNT_STATE = "Order rejected due to the account's status (suspended inactive dup cards)"
    INVALID_WALLET_STATUS = "Order rejected due to the account's wallet's status"
    INSUFFICIENT_FUNDS = "Account has exceeded its exposure limit or available to bet limit"
    LOSS_LIMIT_EXCEEDED = "The account has exceed the self imposed loss limit"
    MARKET_SUSPENDED = "Market is suspended"
    MARKET_NOT_OPEN_FOR_BETTING = "Market is not open for betting. It is either not yet active suspended or closed " \
                                  "awaiting settlement."
    DUPLICATE_TRANSACTION = "Duplicate customer reference data submitted - Please note: There is a time window " \
                            "associated with the de-duplication of duplicate submissions which is 60 second"
    INVALID_ORDER = "Order cannot be accepted by the matcher due to the combination of actions. For example bets " \
                    "being edited are not on the same market or order includes both edits and placement"
    INVALID_MARKET_ID = "Market doesn't exist"
    PERMISSION_DENIED = "Business rules do not allow order to be placed. You are either attempting to place the " \
                        "order using a Delayed Application Key or from a restricted jurisdiction (i.e. USA)"
    DUPLICATE_BETIDS = "duplicate bet ids found"
    NO_ACTION_REQUIRED = "Order hasn't been passed to matcher as system detected there will be no state change"
    SERVICE_UNAVAILABLE = "The requested service is unavailable"
    REJECTED_BY_REGULATOR = "The regulator rejected the order. On the Italian Exchange this error will occur if " \
                            "more than 50 bets are sent in a single placeOrders request."


class StreamingProtocolErrors(Enum):
    """General errors not sent with id linking to specific request (as no request context)
    """
    INVALID_INPUT = 'Failure code returned when an invalid input is provided (could not deserialize the message)'
    TIMEOUT = 'Failure code when a client times out (i.e. too slow sending data)'


class StreamingAuthenticationErrors(Enum):
    """Specific to authentication
    """
    NO_APP_KEY = 'Failure code returned when an application key is not found in the message'
    INVALID_APP_KEY = 'Failure code returned when an invalid application key is received'
    NO_SESSION = 'Failure code returned when a session token is not found in the message'
    INVALID_SESSION_INFORMATION = 'Failure code returned when an invalid session token is received'
    NOT_AUTHORIZED = 'Failure code returned when client is not authorized to perform the operation'
    MAX_CONNECTION_LIMIT_EXCEEDED = 'Failure code returned when a client tries to create more connections than ' \
                                    'allowed to'


class StreamingSubscriptionErrors(Enum):
    """Specific to subscription requests
    """
    SUBSCRIPTION_LIMIT_EXCEEDED = 'Customer tried to subscribe to more markets than allowed to'
    INVALID_CLOCK = 'Failure code returned when an invalid clock is provided on re-subscription (check initialClk / ' \
                    'clk supplied)'


class StreamingGeneralErrors(Enum):
    """General errors which may or may not be linked to specific request id
    """
    UNEXPECTED_ERROR = 'Failure code returned when an internal error occurred on the server'
    CONNECTION_FAILED = 'Failure code used when the client / server connection is terminated'


class StreamingSide(Enum):
    """Some enums are provided in shorthand
    """
    L = 'LAY'
    B = 'BACK'


class StreamingStatus(Enum):
    E = 'EXECUTABLE'
    EC = 'EXECUTION_COMPLETE'


class StreamingPersistenceType(Enum):
    L = 'LAPSE'
    P = 'PERSIST'
    MOC = 'MARKET_ON_CLOSE'


class StreamingOrderType(Enum):
    L = 'LIMIT'
    MOC = 'MARKET_ON_CLOSE'
    LOC = 'LIMIT_ON_CLOSE'


class StreamingRegulatorCode(Enum):
    REG_GGC = 'GIBRALTAR REGULATOR'



import datetime
import unittest

from betfairlightweight.resources.baseresource import BaseResource
from betfairlightweight.streaming.cache import (
    OrderBookCache,
    OrderBookRunner,
    UnmatchedOrder,
    MarketBookCache,
    RunnerBook,
    Available,
)
from betfairlightweight.exceptions import CacheError
from tests import mock
from tests.unit.tools import create_mock_json


class TestAvailable(unittest.TestCase):

    def setUp(self):
        self.prices = [[1, 1.02, 34.45], [0, 1.01, 12]]
        self.available = Available(self.prices, 2)

    def test_init(self):
        assert self.available.prices == self.prices
        assert self.available.deletion_select == 2
        assert self.available.reverse is False

    def test_sort(self):
        self.available.sort()
        assert self.available.prices == self.prices
        assert self.available.serialise == [{'price': 1.01, 'size': 12}, {'price': 1.02, 'size': 34.45}]

    def test_sort_short(self):
        current = [[27, 0.95], [13, 28.01], [1.02, 1157.21]]
        available = Available(current, 1)

        assert available.serialise == [
            {'price': 1.02, 'size': 1157.21}, {'price': 13, 'size': 28.01}, {'price': 27, 'size': 0.95}
        ]

    def test_clear(self):
        self.available.clear()
        assert self.available.prices == []

    def test_update_available_new_update(self):
        # [price, size]
        book_update = [[30, 6.9]]
        current = [[27, 0.95], [13, 28.01], [1.02, 1157.21]]
        expected = [[1.02, 1157.21],  [13, 28.01], [27, 0.95], [30, 6.9]]

        available = Available(current, 1)
        available.update(book_update)
        assert current == expected

        book_update = [[30, 6.9], [1.01, 12]]
        current = [[27, 0.95], [13, 28.01], [1.02, 1157.21]]
        expected = [[1.01, 12], [1.02, 1157.21], [13, 28.01], [27, 0.95], [30, 6.9]]

        available = Available(current, 1)
        available.update(book_update)
        assert current == expected

        # [position, price, size]
        book_update = [[0, 36, 0.57]]
        current = []
        expected = [[0, 36, 0.57]]

        available = Available(current, 2)
        available.update(book_update)
        assert available.prices == expected

    def test_update_available_new_replace(self):
        # [price, size]
        book_update = [[27, 6.9]]
        current = [[27, 0.95], [13, 28.01], [1.02, 1157.21]]
        expected = [[1.02, 1157.21], [13, 28.01], [27, 6.9]]

        available = Available(current, 1)
        available.update(book_update)
        assert current == expected

        # [position, price, size]
        book_update = [[0, 36, 0.57]]
        current = [[0, 36, 10.57], [1, 38, 3.57]]
        expected = [[0, 36, 0.57], [1, 38, 3.57]]

        available = Available(current, 2)
        available.update(book_update)
        assert current == expected

        # tests handling of betfair bug, http://forum.bdp.betfair.com/showthread.php?t=3351
        book_update = [[2, 0, 0], [1, 1.01, 9835.74], [0, 1.02, 1126.22]]
        current = [[1, 1.01, 9835.74], [0, 1.02, 1126.22]]
        expected = [[0, 1.02, 1126.22], [1, 1.01, 9835.74]]

        available = Available(current, 2)
        available.update(book_update)
        assert current == expected

    def test_update_available_new_remove(self):
        book_update = [[27, 0]]
        current = [[27, 0.95], [13, 28.01], [1.02, 1157.21]]
        expected = [[1.02, 1157.21], [13, 28.01]]

        available = Available(current, 1)
        available.update(book_update)
        assert current == expected

        # [position, price, size]
        book_update = [[0, 36, 0], [1, 38, 0], [0, 38, 3.57]]
        current = [[0, 36, 10.57], [1, 38, 3.57]]
        expected = [[0, 38, 3.57]]

        available = Available(current, 2)
        available.update(book_update)
        assert current == expected


class TestMarketBookCache(unittest.TestCase):

    def setUp(self):
        self.market_book_cache = MarketBookCache(**{'marketDefinition': {'runners': {}}})

    def test_error(self):
        with self.assertRaises(CacheError):
            self.market_book_cache = MarketBookCache()

    @mock.patch('betfairlightweight.streaming.cache.MarketBookCache.strip_datetime')
    def test_update_cache_md(self, mock_strip_datetime):
        publish_time = mock.Mock()
        market_change = create_mock_json('tests/resources/streaming_mcm_UPDATE_md.json')
        book_data = market_change.json().get('mc')

        for book in book_data:
            self.market_book_cache.update_cache(book, publish_time)
            mock_strip_datetime.assert_called_with(publish_time)
            assert self.market_book_cache.market_definition == book.get('marketDefinition')

    @mock.patch('betfairlightweight.streaming.cache.MarketBookCache.strip_datetime')
    def test_update_cache_tv(self, mock_strip_datetime):
        publish_time = mock.Mock()
        market_change = create_mock_json('tests/resources/streaming_mcm_UPDATE_tv.json')
        book_data = market_change.json().get('mc')

        for book in book_data:
            self.market_book_cache.update_cache(book, publish_time)
            mock_strip_datetime.assert_called_with(publish_time)
            assert self.market_book_cache.total_matched == book.get('tv')

    # @mock.patch('betfairlightweight.resources.streamingresources.MarketBookCache.strip_datetime')
    # def test_update_cache_rc(self, mock_strip_datetime):
    #     publish_time = mock.Mock()
    #     market_change = create_mock_json('tests/resources/streaming_mcm_UPDATE.json')
    #     book_data = market_change.json().get('mc')
    #
    #     for book in book_data:
    #         self.market_book_cache.update_cache(book, publish_time)
    #         mock_strip_datetime.assert_called_with(publish_time)
    #
    #         assert self.market_book_cache.total_matched == book.get('tv')

    @mock.patch('betfairlightweight.streaming.cache.MarketBookCache.serialise')
    @mock.patch('betfairlightweight.streaming.cache.MarketDefinition')
    @mock.patch('betfairlightweight.streaming.cache.MarketBook')
    def test_create_resource(self, mock_market_book, mock_market_definition, mock_serialise):
        # lightweight
        market_book = self.market_book_cache.create_resource(1234, {}, True)
        assert market_book == mock_serialise
        # not lightweight
        market_book = self.market_book_cache.create_resource(1234, {}, False)
        assert market_book == mock_market_book()

    def test_update_runner_dict(self):
        assert self.market_book_cache.runner_dict == {}

        class Runner:
            def __init__(self, selection_id, name, handicap):
                self.selection_id = selection_id
                self.name = name
                self.handicap = handicap

        (a, b) = (Runner(123, 'a', 1.25), Runner(456, 'b', -0.25))
        self.market_book_cache.runners = [a, b]
        self.market_book_cache._update_runner_dict()
        assert self.market_book_cache.runner_dict == {(123, 1.25): a, (456, -0.25): b}

    def test_init_multiple_rc(self):
        # Initialize data with multiple rc entries for the same selection
        data = {'marketDefinition': {'runners': {}}}
        data['rc'] = [{'atb': [[1.01, 200]], 'id': 13536143}, {'atl': [[1000.0, 200]], 'id': 13536143}]

        market_book_cache = MarketBookCache(**data)

        assert len(market_book_cache.runners) == len(market_book_cache.runner_dict)


class TestRunnerBook(unittest.TestCase):

    def setUp(self):
        self.runner_book = RunnerBook(**{'id': 123})

    def test_update_traded(self):
        self.mock_traded = mock.Mock()
        self.runner_book.traded = self.mock_traded

        self.runner_book.update_traded([])
        self.mock_traded.clear.assert_called_with()

        self.runner_book.update_traded([1, 2])
        self.mock_traded.update.assert_called_with([1, 2])

    def test_serialise_back(self):
        mock_available_to_back = mock.Mock()
        mock_available_to_back.prices = True
        mock_best_available_to_back = mock.Mock()
        mock_best_available_to_back.prices = True
        mock_best_display_available_to_back = mock.Mock()
        mock_best_display_available_to_back.prices = True
        self.runner_book.available_to_back = mock_available_to_back

        assert self.runner_book.serialise_available_to_back() == mock_available_to_back.serialise

        mock_available_to_back.prices = False
        self.runner_book.best_available_to_back = mock_best_available_to_back
        assert self.runner_book.serialise_available_to_back() == mock_best_available_to_back.serialise

        mock_best_available_to_back.prices = False
        self.runner_book.best_display_available_to_back = mock_best_display_available_to_back
        assert self.runner_book.serialise_available_to_back() == mock_best_display_available_to_back.serialise

    def test_serialise_lay(self):
        mock_available_to_lay = mock.Mock()
        mock_available_to_lay.prices = True
        mock_best_available_to_lay = mock.Mock()
        mock_best_available_to_lay.prices = True
        mock_best_display_available_to_lay = mock.Mock()
        mock_best_display_available_to_lay.prices = True
        self.runner_book.available_to_lay = mock_available_to_lay

        assert self.runner_book.serialise_available_to_lay() == mock_available_to_lay.serialise

        mock_available_to_lay.prices = False
        self.runner_book.best_available_to_lay = mock_best_available_to_lay
        assert self.runner_book.serialise_available_to_lay() == mock_best_available_to_lay.serialise

        mock_best_available_to_lay.prices = False
        self.runner_book.best_display_available_to_lay = mock_best_display_available_to_lay
        assert self.runner_book.serialise_available_to_lay() == mock_best_display_available_to_lay.serialise

    def test_empty_serialise(self):
        runner_definition = {
            'bdp': None,
        }
        serialise_d = self.runner_book.serialise(runner_definition)

        ex = serialise_d['ex']
        # all empty lists
        assert all(not ex[a] for a in ex.keys())

        sp = serialise_d['sp']
        # all 'None' or empty lists
        assert all(not sp[a] for a in sp.keys())


class TestOrderBookCache(unittest.TestCase):

    def setUp(self):
        self.order_book_cache = OrderBookCache(**{})
        self.runner = mock.Mock()
        self.runner.selection_id = 10895629
        self.runner.serialise_orders = mock.Mock(return_value=[])
        self.order_book_cache.runners = [self.runner]

    def test_update_cache(self):
        mock_response = create_mock_json('tests/resources/streaming_ocm_UPDATE.json')
        for order_book in mock_response.json().get('oc'):
            self.order_book_cache.update_cache(order_book, 1234)

            for order_changes in order_book.get('orc'):
                # self.runner.matched_lays.update.assert_called_with(order_changes.get('ml', []))
                # self.runner.matched_backs.update.assert_called_with(order_book.get('mb', []))
                self.runner.update_unmatched.assert_called_with(order_changes.get('uo', []))

    @mock.patch('betfairlightweight.streaming.cache.OrderBookRunner')
    def test_update_cache_new(self, mock_order_book_runner):
        self.runner.selection_id = 108956
        mock_response = create_mock_json('tests/resources/streaming_ocm_UPDATE.json')
        for order_book in mock_response.json().get('oc'):
            self.order_book_cache.update_cache(order_book, 1234)

            for order_changes in order_book.get('orc'):
                mock_order_book_runner.assert_called_with(**order_changes)

    @mock.patch('betfairlightweight.streaming.cache.OrderBookCache.serialise')
    @mock.patch('betfairlightweight.streaming.cache.CurrentOrders')
    def test_create_resource(self, mock_current_orders, mock_serialise):
        current_orders = self.order_book_cache.create_resource(123, {}, False)

        assert current_orders == mock_current_orders()

    def test_runner_dict(self):

        class Runner:
            def __init__(self, selection_id, name):
                self.selection_id = selection_id
                self.name = name

        (a, b) = (Runner(123, 'a'), Runner(456, 'b'))
        self.order_book_cache.runners = [a, b]
        assert self.order_book_cache.runner_dict == {123: a, 456: b}

    def test_serialise(self):
        serialised = self.order_book_cache.serialise

        assert serialised == {'currentOrders': [], 'moreAvailable': False}


class TestOrderBookRunner(unittest.TestCase):

    def setUp(self):
        self.order_book_runner = OrderBookRunner(**{'id': 1, 'ml': [], 'mb': [], 'uo': []})


class TestUnmatchedOrder(unittest.TestCase):

    def setUp(self):
        order = {
            'id': 1, 'p': 2, 's': 3, 'side': 'L', 'status': 'E', 'pt': 'L', 'ot': 'L', 'pd': 8, 'sm': 9, 'sr': 10,
            'sl': 11, 'sc': 12, 'sv': 13, 'rfo': 14, 'rfs': 15, 'ld': 16, 'lsrc': 17, 'error': 'test'
        }
        self.unmatched_order = UnmatchedOrder(**order)

    def test_init(self):
        assert self.unmatched_order.bet_id == 1
        assert self.unmatched_order.price == 2
        assert self.unmatched_order.size == 3
        assert self.unmatched_order.side == 'L'
        assert self.unmatched_order.status == 'E'
        assert self.unmatched_order.persistence_type == 'L'
        assert self.unmatched_order.order_type == 'L'
        assert self.unmatched_order.placed_date == BaseResource.strip_datetime(8)
        assert self.unmatched_order.size_matched == 9
        assert self.unmatched_order.size_remaining == 10
        assert self.unmatched_order.size_lapsed == 11
        assert self.unmatched_order.size_cancelled == 12
        assert self.unmatched_order.size_voided == 13
        assert self.unmatched_order.reference_order == 14
        assert self.unmatched_order.reference_strategy == 15
        assert self.unmatched_order.lapsed_date == 16
        assert self.unmatched_order.lapse_status_reason_code == 17

    def test_placed_date_string(self):
        now = datetime.datetime.now()
        self.unmatched_order.placed_date = now
        assert self.unmatched_order.placed_date_string == now.strftime('%Y-%m-%dT%H:%M:%S.%fZ')

    def test_matched_date_string(self):
        now = datetime.datetime.now()
        self.unmatched_order.matched_date = now
        assert self.unmatched_order.matched_date_string == now.strftime('%Y-%m-%dT%H:%M:%S.%fZ')

    def test_serialise(self):
        assert self.unmatched_order.serialise('1.23', 12345, 0.0) == {
            'sizeLapsed': 11, 'persistenceType': 'LAPSE', 'sizeRemaining': 10,
            'placedDate': '1970-01-01T00:00:00.008000Z', 'sizeVoided': 13, 'sizeCancelled': 12, 'betId': 1,
            'customerOrderRef': 14, 'orderType': 'LIMIT', 'marketId': '1.23', 'matchedDate': None, 'side': 'LAY',
            'selectionId': 12345, 'bspLiability': None, 'sizeMatched': 9, 'handicap': 0.0, 'averagePriceMatched': 0.0,
            'status': 'EXECUTABLE', 'customerStrategyRef': 15, 'regulatorCode': None,
            'priceSize': {'price': 2, 'size': 3}
        }



import datetime

from ..compat import (
    basestring,
    integer_types,
    json,
    parse_datetime,
)


class BaseResource(object):
    """Lightweight data structure for resources.
    """

    def __init__(self, **kwargs):
        self.elapsed_time = kwargs.pop('elapsed_time', None)
        now = datetime.datetime.utcnow()
        self._datetime_created = now
        self._datetime_updated = now
        self._data = kwargs

    def json(self):
        return json.dumps(self._data)

    @staticmethod
    def strip_datetime(value):
        """
        Converts value to datetime if string or int.
        """
        if isinstance(value, basestring):
            try:
                return parse_datetime(value)
            except ValueError:
                return
        elif isinstance(value, integer_types):
            try:
                return datetime.datetime.utcfromtimestamp(value / 1e3)
            except (ValueError, OverflowError, OSError):
                return

    def __repr__(self):
        return '<%s>' % self.__class__.__name__

    def __str__(self):
        return self.__class__.__name__



import sys
import datetime


_ver = sys.version_info

#: Python 2.x?
is_py2 = (_ver[0] == 2)

#: Python 3.x?
is_py3 = (_ver[0] == 3)


try:
    from builtins import FileNotFoundError
except ImportError:
    class FileNotFoundError(OSError):
        pass


if is_py2:
    basestring = basestring
    numeric_types = (int, long, float)
    integer_types = (int, long)
elif is_py3:
    basestring = (str, bytes)
    numeric_types = (int, float)
    integer_types = (int,)

# will attempt to use c libraries if installed

try:
    import ujson as json
except ImportError:
    import json

try:
    import ciso8601

    def parse_datetime(datetime_string):
        try:
            return ciso8601.parse_datetime_as_naive(datetime_string)
        except ValueError:
            return
except ImportError:
    def parse_datetime(datetime_string):
        try:
            return datetime.datetime.strptime(datetime_string, "%Y-%m-%dT%H:%M:%S.%fZ")
        except ValueError:
            return



import unittest

from betfairlightweight.apiclient import APIClient


class APIClientInit(unittest.TestCase):

    def test_base_client_init(self):
        client = APIClient('username', 'password', 'app_key')
        assert str(client) == 'APIClient'
        assert repr(client) == '<APIClient [username]>'



from __future__ import print_function

import unittest

from betfairlightweight import resources
from tests.unit.tools import create_mock_json


class ScoreResourcesTest(unittest.TestCase):

    def test_racedetails(self):
        mock_response = create_mock_json('tests/resources/racedetails.json')
        resource = resources.RaceDetails(**mock_response.json())

        assert isinstance(resource, resources.RaceDetails)

    # def test_score(self):
    #     mock_response = create_mock_json('tests/resources/score.json')
    #     resource = resources.Score(**mock_response.json())
    #
    #     assert isinstance(resource, resources.Score)
    #
    # def test_incidents(self):
    #     mock_response = create_mock_json('tests/resources/incidents.json')
    #     resource = resources.Incidents(**mock_response.json())
    #
    #     assert isinstance(resource, resources.Incidents)
    #
    # def test_available_events(self):
    #     mock_response = create_mock_json('tests/resources/availableevents.json')
    #     resource = resources.AvailableEvent(**mock_response.json())
    #
    #     assert isinstance(resource, resources.AvailableEvent)
